/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2015 Josh Blum
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "SoapyHackRF.hpp"
#include <SoapySDR/Logger.hpp>
#include <mutex>
#include <cstddef>

static std::mutex sessionMutex;
static size_t sessionCount = 0;

SoapyHackRFSession::SoapyHackRFSession(void)
{
    std::lock_guard<std::mutex> lock(sessionMutex);

    if (sessionCount == 0)
    {
        int ret = hackrf_init();
        if (ret != HACKRF_SUCCESS)
        {
            SoapySDR::logf(SOAPY_SDR_ERROR, "hackrf_init() failed -- %s", hackrf_error_name(hackrf_error(ret)));
        }
    }
    sessionCount++;
}

SoapyHackRFSession::~SoapyHackRFSession(void)
{
    std::lock_guard<std::mutex> lock(sessionMutex);

    sessionCount--;
    if (sessionCount == 0)
    {
        int ret = hackrf_exit();
        if (ret != HACKRF_SUCCESS)
        {
            SoapySDR::logf(SOAPY_SDR_ERROR, "hackrf_exit() failed -- %s", hackrf_error_name(hackrf_error(ret)));
        }
    }
}
